<?php
// This file is part of Moodle - http://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <http://www.gnu.org/licenses/>.

/**
 * iPay enrolment plugin.
 *
 * This plugin allows you to set up paid courses.
 *
 * @package    enrol_ipay
 * @copyright  iPay Limited
 * @license    http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */

defined('MOODLE_INTERNAL') || die();

/**
 * iPay enrolment plugin implementation.
 * @author  Eugene Venter - based on code by Martin Dougiamas and others
 * @license http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */
class enrol_ipay_plugin extends enrol_plugin {

	public function get_currencies() {
		// See https://www.ipay.com/cgi-bin/webscr?cmd=p/sell/mc/mc_intro-outside,
		// 3-character ISO-4217: https://cms.ipay.com/us/cgi-bin/?cmd=_render-content&content_ID=developer/e_howto_api_currency_codes
		$codes = array('USD', 'KES', 'UGX', 'TZS', 'XOF');
		$currencies = array();
		foreach ($codes as $c) {
			$currencies[$c] = new lang_string($c, 'core_currencies');
		}

		return $currencies;
	}

	public function get_countries() {
		return array(
			'ke' => "Kenya",
			'tz' => "Tanzania",
			'ug' => "Uganda",
			'tg' => "Togo",
		);
	}

	public function get_yesno() {
		return array(
			'0' => 'No',
			'1' => 'Yes',
		);
	}

	/**
	 * Returns optional enrolment information icons.
	 *
	 * This is used in course list for quick overview of enrolment options.
	 *
	 * We are not using single instance parameter because sometimes
	 * we might want to prevent icon repetition when multiple instances
	 * of one type exist. One instance may also produce several icons.
	 *
	 * @param array $instances all enrol instances of this type in one course
	 * @return array of pix_icon
	 */
	public function get_info_icons(array $instances) {
		$found = false;
		foreach ($instances as $instance) {
			if ($instance->enrolstartdate != 0 && $instance->enrolstartdate > time()) {
				continue;
			}
			if ($instance->enrolenddate != 0 && $instance->enrolenddate < time()) {
				continue;
			}
			$found = true;
			break;
		}
		if ($found) {
			return array(new pix_icon('icon', get_string('pluginname', 'enrol_ipay'), 'enrol_ipay'));
		}
		return array();
	}

	public function roles_protected() {
		// users with role assign cap may tweak the roles later
		return false;
	}

	public function allow_unenrol(stdClass $instance) {
		// users with unenrol cap may unenrol other users manually - requires enrol/ipay:unenrol
		return true;
	}

	public function allow_manage(stdClass $instance) {
		// users with manage cap may tweak period and status - requires enrol/ipay:manage
		return true;
	}

	public function show_enrolme_link(stdClass $instance) {
		return ($instance->status == ENROL_INSTANCE_ENABLED);
	}

	/**
	 * Returns true if the user can add a new instance in this course.
	 * @param int $courseid
	 * @return boolean
	 */
	public function can_add_instance($courseid) {
		$context = context_course::instance($courseid, MUST_EXIST);

		if (!has_capability('moodle/course:enrolconfig', $context) or !has_capability('enrol/ipay:config', $context)) {
			return false;
		}

		// multiple instances supported - different cost for different roles
		return true;
	}

	/**
	 * We are a good plugin and don't invent our own UI/validation code path.
	 *
	 * @return boolean
	 */
	public function use_standard_editing_ui() {
		return true;
	}

	/**
	 * Add new instance of enrol plugin.
	 * @param object $course
	 * @param array $fields instance fields
	 * @return int id of new instance, null if can not be created
	 */
	public function add_instance($course, array $fields = null) {
		if ($fields && !empty($fields['cost'])) {
			$fields['cost'] = unformat_float($fields['cost']);
		}
		return parent::add_instance($course, $fields);
	}

	/**
	 * Update instance of enrol plugin.
	 * @param stdClass $instance
	 * @param stdClass $data modified instance fields
	 * @return boolean
	 */
	public function update_instance($instance, $data) {
		if ($data) {
			$data->cost = unformat_float($data->cost);
		}
		return parent::update_instance($instance, $data);
	}

	/**
	 * Creates course enrol form, checks if form submitted
	 * and enrols user if necessary. It can also redirect.
	 *
	 * @param stdClass $instance
	 * @return string html text, usually a form in a text box
	 */
	function enrol_page_hook(stdClass $instance) {
		global $CFG, $USER, $OUTPUT, $PAGE, $DB;

		$instancename = $this->get_instance_name($instance);

		ob_start();

		if ($DB->record_exists('user_enrolments', array('userid' => $USER->id, 'enrolid' => $instance->id))) {
			return ob_get_clean();
		}

		if ($instance->enrolstartdate != 0 && $instance->enrolstartdate > time()) {
			return ob_get_clean();
		}

		if ($instance->enrolenddate != 0 && $instance->enrolenddate < time()) {
			return ob_get_clean();
		}

		$course = $DB->get_record('course', array('id' => $instance->courseid));
		$context = context_course::instance($course->id);

		$shortname = format_string($course->shortname, true, array('context' => $context));
		$strloginto = get_string("loginto", "", $shortname);
		$strcourses = get_string("courses");

		// Pass $view=true to filter hidden caps if the user cannot see them
		if ($users = get_users_by_capability($context, 'moodle/course:update', 'u.*', 'u.id ASC',
			'', '', '', '', false, true)) {
			$users = sort_by_roleassignment_authority($users, $context);
			$teacher = array_shift($users);
		} else {
			$teacher = false;
		}

		if ((float) $instance->cost <= 0) {
			$cost = (float) $this->get_config('cost');
		} else {
			$cost = (float) $instance->cost;
		}

		if (abs($cost) < 0.01) {
			// no cost, other enrolment methods (instances) should be used
			echo '<p>' . get_string('nocost', 'enrol_ipay') . '</p>';
		} else {

			// Calculate localised and "." cost, make sure we send iPay the same value,
			// please note iPay expects amount with 2 decimal places and "." separator.
			$localisedcost = format_float($cost, 2, true);
			$cost = format_float($cost, 2, false);

			if (isguestuser()) {
				// force login only for guest user, not real users with guest role
				$wwwroot = $CFG->wwwroot;
				echo '<div class="mdl-align"><p>' . get_string('paymentrequired') . '</p>';
				echo '<p><b>' . get_string('cost') . ": $instance->currency $localisedcost" . '</b></p>';
				echo '<p><a href="' . $wwwroot . '/login/">' . get_string('loginsite') . '</a></p>';
				echo '</div>';
			} else {

				$live = $this->get_config('sandbox');
				if ($live == "1") {
					$live = "0";
				} else {
					$live = "1";
				}

				$mer = $this->get_config('merchantname');

				$tel = "";

				if (!empty($USER->phone1)) {
					$tel = $USER->phone1;
				} elseif (!empty($USER->phone2)) {
					$tel = $USER->phone2;
				} else {
					$tel = '0700000000';
				}
				$tel = str_replace(array(' ', '<', '>', '&', '{', '}', '*', "+", '!', '@', '#', "$", '%', '^', '&'), "", $tel);

				$eml = $USER->email;
				$vid = $this->get_config('vendorid');
				$oid = $USER->id . "-" . $course->id . "-" . $instance->id;
				$inv = $oid;
				$p1 = '';
				$p2 = '';
				$p3 = '';
				$p4 = '';

				$curr = $instance->currency;

				$ttl = $cost;
				$crl = '0';
				$cst = '1';
				$cbk = $CFG->wwwroot . "/enrol/ipay/ipn.php";
				$hsh = $this->get_config('hashkey');
				$autopay = $this->get_config('autopay');

				// Identifier and mandatory channels
				$channel = "Moodle";
				$vooma   = "1";
				$bonga   = "1";

				$datastring = $live . $oid . $inv . $ttl . $tel . $eml . $vid . $curr . $p1 . $p2 . $p3 . $p4 . $cbk . $cst . $crl;
				$hash = hash_hmac('sha1', $datastring, $hsh);
				// //Sanitise some fields before building the iPay form
				// $coursefullname = format_string($course->fullname, true, array('context' => $context));
				// $courseshortname = $shortname;
				// $userfullname = fullname($USER);
				// $userfirstname = $USER->firstname;
				// $userlastname = $USER->lastname;
				// $useraddress = $USER->address;
				// $usercity = $USER->city;
				// $this->get_instance_name($instance);

				$ipayurl = "";

				switch ($this->get_config('country')) {
				case 'ke':
					$ipayurl = "https://payments.ipayafrica.com/v3/ke";
					break;
				case 'tz':
					$ipayurl = "https://payments.elipa.co.tz/v3/tz";
					break;
				case 'ug':
					$ipayurl = "https://payments.elipa.co.ug/v3/ug";
					break;
				case 'tg':
					$ipayurl = "https://payments.elipa.tg/v1/v3/index.php/togo";
					break;

				default:
					exit();
				}

				include $CFG->dirroot . '/enrol/ipay/enrol.html';
			}

		}

		return $OUTPUT->box(ob_get_clean());
	}

	/**
	 * Restore instance and map settings.
	 *
	 * @param restore_enrolments_structure_step $step
	 * @param stdClass $data
	 * @param stdClass $course
	 * @param int $oldid
	 */
	public function restore_instance(restore_enrolments_structure_step $step, stdClass $data, $course, $oldid) {
		global $DB;
		if ($step->get_task()->get_target() == backup::TARGET_NEW_COURSE) {
			$merge = false;
		} else {
			$merge = array(
				'courseid' => $data->courseid,
				'enrol' => $this->get_name(),
				'roleid' => $data->roleid,
				'cost' => $data->cost,
				'currency' => $data->currency,
			);
		}
		if ($merge and $instances = $DB->get_records('enrol', $merge, 'id')) {
			$instance = reset($instances);
			$instanceid = $instance->id;
		} else {
			$instanceid = $this->add_instance($course, (array) $data);
		}
		$step->set_mapping('enrol', $oldid, $instanceid);
	}

	/**
	 * Restore user enrolment.
	 *
	 * @param restore_enrolments_structure_step $step
	 * @param stdClass $data
	 * @param stdClass $instance
	 * @param int $oldinstancestatus
	 * @param int $userid
	 */
	public function restore_user_enrolment(restore_enrolments_structure_step $step, $data, $instance, $userid, $oldinstancestatus) {
		$this->enrol_user($instance, $userid, null, $data->timestart, $data->timeend, $data->status);
	}

	/**
	 * Return an array of valid options for the status.
	 *
	 * @return array
	 */
	protected function get_status_options() {
		$options = array(ENROL_INSTANCE_ENABLED => get_string('yes'),
			ENROL_INSTANCE_DISABLED => get_string('no'));
		return $options;
	}

	/**
	 * Return an array of valid options for the roleid.
	 *
	 * @param stdClass $instance
	 * @param context $context
	 * @return array
	 */
	protected function get_roleid_options($instance, $context) {
		if ($instance->id) {
			$roles = get_default_enrol_roles($context, $instance->roleid);
		} else {
			$roles = get_default_enrol_roles($context, $this->get_config('roleid'));
		}
		return $roles;
	}

	/**
	 * Add elements to the edit instance form.
	 *
	 * @param stdClass $instance
	 * @param MoodleQuickForm $mform
	 * @param context $context
	 * @return bool
	 */
	public function edit_instance_form($instance, MoodleQuickForm $mform, $context) {

		$mform->addElement('text', 'name', get_string('custominstancename', 'enrol'));
		$mform->setType('name', PARAM_TEXT);

		$options = $this->get_status_options();
		$mform->addElement('select', 'status', get_string('status', 'enrol_ipay'), $options);
		$mform->setDefault('status', $this->get_config('status'));

		$mform->addElement('text', 'cost', get_string('cost', 'enrol_ipay'), array('size' => 4));
		$mform->setType('cost', PARAM_RAW);
		$mform->setDefault('cost', format_float($this->get_config('cost'), 2, true));

		$ipaycurrencies = $this->get_currencies();
		$mform->addElement('select', 'currency', get_string('currency', 'enrol_ipay'), $ipaycurrencies);
		$mform->setDefault('currency', $this->get_config('currency'));

		$roles = $this->get_roleid_options($instance, $context);
		$mform->addElement('select', 'roleid', get_string('assignrole', 'enrol_ipay'), $roles);
		$mform->setDefault('roleid', $this->get_config('roleid'));

		$options = array('optional' => true, 'defaultunit' => 86400);
		$mform->addElement('duration', 'enrolperiod', get_string('enrolperiod', 'enrol_ipay'), $options);
		$mform->setDefault('enrolperiod', $this->get_config('enrolperiod'));
		$mform->addHelpButton('enrolperiod', 'enrolperiod', 'enrol_ipay');

		$options = array('optional' => true);
		$mform->addElement('date_time_selector', 'enrolstartdate', get_string('enrolstartdate', 'enrol_ipay'), $options);
		$mform->setDefault('enrolstartdate', 0);
		$mform->addHelpButton('enrolstartdate', 'enrolstartdate', 'enrol_ipay');

		$options = array('optional' => true);
		$mform->addElement('date_time_selector', 'enrolenddate', get_string('enrolenddate', 'enrol_ipay'), $options);
		$mform->setDefault('enrolenddate', 0);
		$mform->addHelpButton('enrolenddate', 'enrolenddate', 'enrol_ipay');

		if (enrol_accessing_via_instance($instance)) {
			$warningtext = get_string('instanceeditselfwarningtext', 'core_enrol');
			$mform->addElement('static', 'selfwarn', get_string('instanceeditselfwarning', 'core_enrol'), $warningtext);
		}
	}

	/**
	 * Perform custom validation of the data used to edit the instance.
	 *
	 * @param array $data array of ("fieldname"=>value) of submitted data
	 * @param array $files array of uploaded files "element_name"=>tmp_file_path
	 * @param object $instance The instance loaded from the DB
	 * @param context $context The context of the instance we are editing
	 * @return array of "element_name"=>"error_description" if there are errors,
	 *         or an empty array if everything is OK.
	 * @return void
	 */
	public function edit_instance_validation($data, $files, $instance, $context) {
		$errors = array();

		if (!empty($data['enrolenddate']) and $data['enrolenddate'] < $data['enrolstartdate']) {
			$errors['enrolenddate'] = get_string('enrolenddaterror', 'enrol_ipay');
		}

		$cost = str_replace(get_string('decsep', 'langconfig'), '.', $data['cost']);
		if (!is_numeric($cost)) {
			$errors['cost'] = get_string('costerror', 'enrol_ipay');
		}

		$validstatus = array_keys($this->get_status_options());
		$validcurrency = array_keys($this->get_currencies());
		$validroles = array_keys($this->get_roleid_options($instance, $context));
		$tovalidate = array(
			'name' => PARAM_TEXT,
			'status' => $validstatus,
			'currency' => $validcurrency,
			'roleid' => $validroles,
			'enrolperiod' => PARAM_INT,
			'enrolstartdate' => PARAM_INT,
			'enrolenddate' => PARAM_INT,
		);

		$typeerrors = $this->validate_param_types($data, $tovalidate);
		$errors = array_merge($errors, $typeerrors);

		return $errors;
	}

	/**
	 * Execute synchronisation.
	 * @param progress_trace $trace
	 * @return int exit code, 0 means ok
	 */
	public function sync(progress_trace $trace) {
		$this->process_expirations($trace);
		return 0;
	}

	/**
	 * Is it possible to delete enrol instance via standard UI?
	 *
	 * @param stdClass $instance
	 * @return bool
	 */
	public function can_delete_instance($instance) {
		$context = context_course::instance($instance->courseid);
		return has_capability('enrol/ipay:config', $context);
	}

	/**
	 * Is it possible to hide/show enrol instance via standard UI?
	 *
	 * @param stdClass $instance
	 * @return bool
	 */
	public function can_hide_show_instance($instance) {
		$context = context_course::instance($instance->courseid);
		return has_capability('enrol/ipay:config', $context);
	}
}
